/*********************************************\
* NO_DUPE.c   rev 1.0                         *
* Adapted from: C USERS JOURNAL Feb 91 pp64   *
* by Chuck Maier  Feb 1991                    *
* compiled for Lattice with:                  *
*                                             *
*          LC -L -md no_dupe                  *
\*********************************************/


#include <dos.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>


#define VERSION "NO_DUP rev 1.0"
#define ALL_FILES 0xFFFF
#define MAX_LINE 80
#define MAX_PATH 80
#define MAX_FLEN 13
#define MAX_FILES 15000
#define SEPARATOR "\n----------------------------------------------"

typedef enum{ FALSE, TRUE } Flag;

int line_count, pause_flag;

typedef struct DirEntry {
    char *path;
    char name[13];
    char attrib;
    long time;
    long size;
    Flag dir_processed;
} DirEntry;

typedef struct DirList {
    char pathname[MAX_PATH];
    Flag dir_processed;
    struct DirList *prev;
    struct DirList *next;
} DirList;

typedef struct {
    Flag dir;
    Flag file;
    unsigned int file_count;
    unsigned long total_file_size;
}GlobalOpt;

typedef char *PtrList;

DirEntry *get_direntry (char *);
DirList *make_path (char *, char *, DirList *);
DirList *get_path (DirList *);
int name_comp(const void *,const void *);
char *datestr (long d, char *buf);
char *timestr (long t, char *buf);
void fprint_direntry ( DirEntry *);

Flag first_call = TRUE;
GlobalOpt globalopt = {FALSE,FALSE,0,0L};



int main(argc,argv)
int argc;
char *argv[];
{
    DirEntry *current_file, *next_file;
    DirList *path, *root, *new_path;
    char current_path[MAX_PATH],file_spec[MAX_FLEN];
    PtrList *ptrbase,*ptrlist;
    Flag duplicate = FALSE;
    char ww;
    int a;

    for (a=argc;a>1;a--){
        ww=argv[a-1][0];
        if(ww == 'f' || ww== 'F'){
            globalopt.file=TRUE;
        } else if( ww=='d' || ww=='D'){
            globalopt.dir=TRUE;
        } else if(ww=='p' || ww=='P'){
            line_count=0;
            pause_flag=1;
        } else {
            printf("Invalid command line switch: %s\n",
             argv[a-1]);
            printf("\nValid arguments are: F=filelist D=Directories P=pause");
            printf("\n Each argument must be separated by spaces");
        }
    }
    strcpy(file_spec,"*.*");
    current_file=NULL;
    (void)getcwd(current_path,MAX_PATH);
    printf(VERSION);
    printf("\nProcessing %s",current_path);
    path=make_path(current_path,"",NULL);
    new_path=root=path ;
    if((ptrbase=malloc(sizeof(PtrList) * MAX_FILES))==NULL){
        printf("\nInsufficient near memory!!");
        printf("\nProcessed %d files occupying %lu bytes.",
        globalopt.file_count,globalopt.total_file_size);
        exit(1);
    }
    ptrlist=ptrbase;
    while((path = get_path(root))!=NULL){
        strcpy(current_path,path->pathname);
        strcat(current_path,file_spec);
        first_call=TRUE;
        while((current_file=get_direntry(current_path))!=NULL){
            current_file->path=path->pathname;
            if((current_file->attrib & 0x10) && current_file->name[0]=='.') continue;
            if(current_file->attrib & 0x10) {
                new_path=make_path(path->pathname,current_file->name,new_path);
            } else {
                ++globalopt.file_count;
                globalopt.total_file_size += current_file->size;
                if(globalopt.file_count==MAX_FILES){
                    printf("\nTOO MANY FILES... max=%u\n",MAX_FILES);
                    exit(1);
                }
                *ptrlist=(char *) current_file;
                ++ptrlist;
            }
        }
        path->dir_processed=TRUE;
    }
    *ptrlist=NULL;
    printf("\nProcessed %d files occupying %lu bytes.",
      globalopt.file_count,globalopt.total_file_size);
    qsort((void *)ptrbase,(size_t)globalopt.file_count,
      (size_t)sizeof(PtrList),name_comp);
    if(globalopt.file){  /* optional file list */
        printf(SEPARATOR);
        line_count++;
        printf("\nListing of files below %s",root);
        for(ptrlist=ptrbase;*ptrlist;ptrlist++){
            fprint_direntry((DirEntry *)*ptrlist);
        }
    }
    printf(SEPARATOR);
    line_count++;
    printf("\nDuplicate files below %s",root);
    for(ptrlist=ptrbase,duplicate=FALSE;*ptrlist;ptrlist++){
        current_file=(DirEntry *)*ptrlist;
        next_file=(DirEntry *) *(ptrlist+1);
        if(!strcmp(current_file->name,next_file->name)){
            duplicate=TRUE;
            fprint_direntry(current_file);
        } else {
            if(duplicate==TRUE){
                /* print the last file in a group or we will miss it */
                fprint_direntry(current_file);
                printf(SEPARATOR);
                line_count++;
                duplicate=FALSE;
            }
        }
    }
}

/**********************/
DirEntry * get_direntry(upath)
char *upath;
{
    int stat;
    extern int errno;
    static struct FILEINFO fifo;
    DirEntry *new_entry;
    if(first_call==TRUE){
        stat = dfind(&fifo,upath,0x10);
        first_call=FALSE;
    } else {
        stat = dnext(&fifo);
    }
    if(stat == -1){
       return(NULL);
    }
    if((new_entry = malloc(sizeof(DirEntry)))==NULL){
        printf("\nInsufficient far memory!!");
        printf("\nProcessed %d files occupying %lu bytes.",
        globalopt.file_count,globalopt.total_file_size);

        exit(1);
    }
    strcpy(new_entry->name,fifo.name);
    new_entry->attrib = fifo.attr;
    if(new_entry->attrib & 0x10){
        new_entry->dir_processed = FALSE;
    }
    new_entry->time = fifo.time;
    new_entry->size = fifo.size;
    return(new_entry);
}

/*************************/
DirList * make_path(basedir, subdir, prev)
char *basedir, *subdir;
DirList *prev;
{
    DirList *new_dir;
    if((new_dir = malloc(sizeof (DirList)))==NULL){
        printf("\nINSUFFICIENT near memory");
        printf("\nProcessed %d files occupying %lu bytes.",
        globalopt.file_count,globalopt.total_file_size);

        exit(1);
    }
    strcpy(new_dir->pathname,basedir);
    strcat(new_dir->pathname,subdir);
    if(*(new_dir->pathname + strlen(new_dir->pathname)-1)!='\\')
        strcat(new_dir->pathname,"\\");
    if(strcmp(subdir,".")&&strcmp(subdir,"..")){
        /* only REAL directories meet the conditions */
        new_dir->dir_processed=FALSE;
        if(globalopt.dir){
            printf("\nDirectory: %s",new_dir->pathname);
        }
    } else {
        new_dir->dir_processed = TRUE;
    }
    if(prev){
        new_dir->prev=prev;
        new_dir->next=NULL;
        prev->next=new_dir;
    } else {
        new_dir->prev=NULL;
        new_dir->next=NULL;
    }
    return(new_dir);
}

/***********************/
DirList * get_path(root)
DirList *root;
{
    DirList *path;

    path=root;
    while(path->next &&(path->dir_processed == TRUE)){
        path=path->next;
    }
    if(path->dir_processed==TRUE){
        return(NULL);
    } else {
        return(path);
    }
}

/*********************/
int name_comp(p1, p2)
const void *p1, *p2;
{
    PtrList *ptr1,*ptr2;
    DirEntry *f1, *f2;
    int status;

    ptr1 = (PtrList *) p1;
    ptr2 = (PtrList *) p2;
    f1 = (DirEntry *) *ptr1;
    f2 = (DirEntry *) *ptr2;

    if(status=strcmp(f1->name,f2->name)) return(status);
    else if(status=f1->time - f2->time) return(status);
    else
    return(strcmp(f1->path,f2->path));
}

/***********************/
void fprint_direntry(cf)
DirEntry *cf;
{
    char date_buff[20],time_buff[20];
    long t;
    char tt[50];
    char *y;
    char ch;
    char out[]={"Enter any key..'C'=Continous "};
    int a;

    t = cf->time;
    y=ctime(&t);
    strcpy(tt,y);
    tt[19]=0x00;
    strcpy(time_buff,&tt[11]);

    y=ctime(&t);
    strcpy(tt,y);
    tt[10] = tt[24] = 0x00;
    sprintf(date_buff,"%s %s",&tt[4],&tt[20]);

    printf("\n%-14s %6ld %c%c %s %s %s", cf->name, cf->size,
      (cf->attrib & 0x01)? 'R':'.',
      (cf->attrib & 0x02)? 'A':'.',
      date_buff,time_buff, cf->path);

   line_count++;
   if((line_count > 19) && pause_flag ){
/*
      printf("\n%s",out);
*/
      ch = getch();
      if(ch=='c' || ch=='C') pause_flag=0;
      line_count=0;
/*
      for(a=0;a<strlen(out);a++){
        printf("\b");
      }
*/
   }
}





